%{
/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

/* This is just a very simple lexical analizer for F90, keywords are reserved
and source form is completely free.
1) ';' also signals end of statement.
2) ^[*c] and ! comments are taken into account. (note that ^[*c] comment
   conflicts with "&\n" continuation.)
3) Deals with both free and fixed source form continuation lines. But
   the continuation cannot break a token...
*/

#include "extern.h"
#include "yacc.h"

#define YY_USER_ACTION zzleng=yyleng;zztext=yytext;

#undef YY_INPUT
#define YY_INPUT(buf,result,max_size) form_input(buf,&result,max_size,yyin)

char *zztext;
int zzlineno = 0, zzleng;
char *lex_last_name;

static int form_stmt_lines;  /* the # of lines of the last statement */
%}

W	[ \t]*

	/* 302 */

Digit		[0-9]
Letter		[a-z]
AlphanumericCharacter	({Letter}|{Digit}|"_"|"$")

	/* 304 */

Name	{Letter}{AlphanumericCharacter}*

	/* 401 */

SignedDigitString	[+-]?{DigitString}

	/* 402 */

DigitString	{Digit}+

	/* 404 */

IntLiteralConstant	{DigitString}{KindParam}?

	/* 405 */

KindParam	("_"({DigitString}|{Name}))

	/* 408 */

BinaryConstant	b(\'[01]+\'|\"[01]+\")

	/* 409 */

OctalConstant	o(\'[0-7]+\'|\"[0-7]+\")

	/* 410 */

HexConstant	z(\'[0-9a-f]+\'|\"[0-9a-f]+\")

	/* 413 */

RealLiteralConstant	({Significand}([e]{SignedDigitString})?|{DigitString}[e]{SignedDigitString}){KindParam}?

DoubleLiteralConstant	({Significand}([d]{SignedDigitString})?|{DigitString}[d]{SignedDigitString}){KindParam}?

	/* 414 */

Significand	(({DigitString}"."{DigitString}?)|("."{DigitString}))

	/* 421 */

LogicalLiteralConstant	(".true."|".false."){KindParam}?

	/* ~ 311, 704, 724 */

DefinedOperator	"."{Letter}+"."

%%

{IntLiteralConstant} {
	Node *node = make (INTEGER_CONSTANT);

	kind_check ();
	node->com.str = U_str_copy (yytext);
	node->com.typ = typ_malloc ();
	node->com.typ->etype = Etype_INTEGER_0;
	yylval.node = node;
	return (INTEGER_CONSTANT);
}

{BinaryConstant} {
	kind_check ();
	return (BINARY_CONSTANT);
}

{OctalConstant} {
	kind_check ();
	return (OCTAL_CONSTANT);
}

{HexConstant} {
	kind_check ();
	return (HEX_CONSTANT);
}

{RealLiteralConstant} {
	int c;
	char *str = U_str_copy (yytext);

	c = input();
	unput(c);

	if (isalpha(c)) {  /* catches .[a-z]+. */
		U_free (str);
		REJECT;
	}

	{
	Node *node = make (REAL_CONSTANT);

	kind_check ();

	node->com.str = str;
	node->com.typ = typ_malloc ();
	node->com.typ->etype = Etype_REAL_0;
	yylval.node = node;
	}

	return (REAL_CONSTANT);
}

{DoubleLiteralConstant} {
	int c;
	char *str = U_str_copy (yytext);

	c = input();
	unput(c);

	if (isalpha(c)) {  /* catches .[a-z]+. */
		U_free (str);
		REJECT;
	}

	{
	Node *node = make (REAL_CONSTANT);

	kind_check ();

	node->com.str = str;
	node->com.typ = typ_malloc ();
	node->com.typ->etype = Etype_REAL_1;
	yylval.node = node;
	}

	return (REAL_CONSTANT);
}

{LogicalLiteralConstant} {
	{
	Node *node = make (LOGICAL_CONSTANT);

	kind_check ();

	node->com.str = U_str_copy (yytext);
	node->com.typ = typ_malloc ();
	node->com.typ->etype = Etype_LOGICAL_0;
	yylval.node = node;
	}
	return (LOGICAL_CONSTANT);
}

"{"			return (LPEQI);
"~"			return (CHAR_SELECTOR_COMMA);

"("{W}"/"	return (LDAC);
"/"{W}")"	return (RDAC);
"[" {
	stand ("'[' as array constructor delimiter");
	return (LDAC);
}
"]" {
	stand ("']' as array constructor delimiter");
	return (RDAC);
}

"="		return ('=');
"+"		return ('+');
"-"		return ('-');
"*"		return ('*');
"/"		return ('/');
"("		return ('(');
")"		return (')');
","		return (',');
"."		return ('.');
":"		return (':');
"%"		return ('%');
"?"		return ('?');
"$"		return ('$');

"**"		return (POWER_OP);

"//"		return (CONCAT_OP);

".eq."		|
"=="		return (REL_OP_EQ);
".ne."		|
"\\="		return (REL_OP_NE);
"<>" {
	stand ("relational operator '<>'");
	return (REL_OP_NE);
}
".lt."		|
"<"			return (REL_OP_LT);
".le."		|
"<="		return (REL_OP_LE);
".gt."		|
">"			return (REL_OP_GT);
".ge."		|
">="		return (REL_OP_GE);

"=>"		return (POINT_OP);

".not."		return (NOT_OP);
".and."		return (AND_OP);
".or."		return (OR_OP);
".xor."		return (XOR_OP);  /* this is an extension */
".eqv."		return (EQUIV_OP_T);
".neqv."	return (EQUIV_OP_F);

{DefinedOperator} {
	process_string('.');
	return (DEFINED_UNARY_OP);
}

"#s" {
	Node *node = make (CHARACTER_CONSTANT);
	char *str = strings_list_fun (strings_current++);
	char *string_fix ();

	zztext = str;
	node->com.str = string_fix (str);

	node->com.typ = typ_malloc ();
	node->com.typ->etype = Etype_CHARACTER_0;

	yylval.node = node;

	return (CHARACTER_CONSTANT);
}

"#" {
	int type = key_list_fun (key_current++);

	zztext = node_Symbol (type);

	if (type == FORMAT) {
		char c, *str;
		int i, num = -1, len;
		Node *node = make (FORMAT);

		strcpy (serr, "FORMAT (");

		while ((c = input ()) != '(' && c != '{')
			;
	
		for (i = 8;; i++) {
			c = input();
			serr[i] = c;

			if (c == '#') {
				c = input ();

				str = strings_list_fun (strings_current++);
				len = strlen (str);
				strcpy (serr + i, str);
				i = i + len - 1;
				continue;
			}

			switch (c) {

			case '\n':
				unput (c);
				fatal ("unmatched parenthesis in FORMAT");
			case ')':
				num++;
				if (num == 0) {
					serr[i+1] = '\0';
					node->com.str = U_str_copy (serr);
					yylval.node = node;
					return (FORMAT);
				}
				break;

			case '(':
				num--;
				break;
			}
		} 
	}

	return type;
}

@{Name} {
	Node *node;

	if (flag_externals == 0)
		REJECT;

	node = make (NAME);

	node->com.str= U_str_copy (yytext);
	yylval.node = node;

	lex_last_name = node->com.str;

	return (NAME);
}

_{Name} {
	Node *node;

	node = make (NAME);

	sprintf (serr, "name '%s' starts with an underscore", yytext);
	if (flag_standard)
		stand (serr);
	else
		warning (serr);

	node->com.str= U_str_copy (yytext);
	yylval.node = node;

	lex_last_name = node->com.str;

	return (NAME);
}

{Name} {
	Node *node = make (NAME);

	node->com.str= U_str_copy (yytext);
	yylval.node = node;

	lex_last_name = node->com.str;

	return (NAME);
}

;	{
/*!!! should add code that checks if there is a label afterwards ==> stand () */
	BEGIN 0;
	return (EOS);
}

\n	{
	int c, ilen = 0;
	char len [10];

	BEGIN 0;

	while ((c = input ()) != '\n')
		len [ilen++] = c;
	len [ilen] = '\0';

	form_stmt_lines = atoi (len);
	zzlineno = zzlineno + form_stmt_lines;

	if (flag_verbose)
		ui_display (zzlineno);

	return (EOS);
}

[ \t\f\r\v]	;  /*!!! give stand for non-stardard characters */

. {
	sprintf (serr, "'%c' ('\\%3o'): illegal character", yytext[0], yytext[0]);
	fatal (serr);
}

%%

static char *string_fix (str)
char *str;
{
	char *string, *temp0, *temp;

	temp0 = temp = (char*) U_malloc (2 * strlen (str));

	if (*str == '\"') {
		*temp++ = '\'';

		for (;;)
			switch (*++str) {
			case '\'':  /* ' --> '' */
				*temp++ = '\'';
				*temp++ = '\'';
				break;
			case '"':
				if (*(str+1) == '"') {  /* "" --> " */
					*temp++ = '"';
					str++;
				}
				else {
					*temp++ = '\'';
					goto eos;
				}
				break;
			default:
				*temp++ = *str;
			}

eos:
		*temp = '\0';
		string = U_str_copy (temp0);
		U_free (temp0);
	}
	else {
		string = U_str_copy (str);
	}

	return string;
}

static process_string(c)
char c;
{
	char *from = yytext+1, *to = yytext, *end = yytext+zzleng-1;

	zzleng = zzleng-2;

	while (from < end) {
		if (*from == c) {
			from++;
			zzleng--;
		}
		*to++ = *from++;
	}

	*to = '\0';
}

real_zzlineno ()
{
    if (yytext != (char*) 0 && yytext[0] == '\n')
        return zzlineno + 1 - form_stmt_lines;
    else
        return zzlineno + 1;
}

/************************* I N C L U D E   F I L E S ****************************/

int num_yywrap = 0;
static FILE *current_fd = (FILE*) 0;

typedef struct Incfile {
	YY_BUFFER_STATE state;
	char *name;
	int lineno;
	char *old_line;
	int old_len;
	FILE *fd;
} Incfile;

static U_DEFINE_STACK (sinclude, Incfile*, 10)

static Incfile *incfile_malloc_c (state, name, lineno, fd)
YY_BUFFER_STATE state;
char *name;
int lineno;
FILE *fd;
{
	Incfile *incfile;

	incfile = (Incfile*) U_malloc (sizeof (Incfile));
	incfile->state = state;
	incfile->name = name;
	incfile->lineno = lineno;
	incfile->old_line = form_old_line;  /* stores the current ones */
	incfile->old_len = form_old_len;
	form_old_len = -1;
	incfile->fd = fd;

	return incfile;
}

static FILE *open_file (file)
char *file;
{
	FILE *fp;

	if ((fp = fopen (file, "r")) == (FILE*) 0) {
		sprintf (serr, "include file \"%s\"", file);
		comerr (serr);
		fp = (FILE*) 0;
	}

	return fp;
}

static FILE *open_file_include (file)
char *file;
{
	FILE *fp;
	char *include_path, *try;
	int i;

	if ((fp = fopen (file, "r")) != (FILE*) 0)
		return fp;

	if (file [0] != '/') {
		for (i = 0; i < include_path_top_fun (); i++) {
			include_path = include_path_fun (i);

			try = (char*) U_malloc (strlen (file) + strlen (include_path) + 2);
			strcpy (try, include_path);
			strcat (try, "/");
			strcat (try, file);

			fp = fopen (try, "r");
			U_free (try);

			if (fp != (FILE*) 0)
				return fp;
		}
	}

	sprintf (serr, "include file \"%s\"", file);
	comerr (serr);
	return (FILE*) 0;
}

static FILE *open_file_from_user (file)
char *file;
{
	FILE *fp;

	if ((fp = fopen (file, "r")) == (FILE*) 0) {
		sprintf (serr, "f90: \"%s\"", file);
		comerrnl (serr);
		exit (1);
	}

	return fp;
}

include_init ()
{
	char *intrin;
	FILE *fd;

	if (flag_cmf)
		intrin = "lib/intrincm.itf";
	else
		intrin = "lib/intrin.itf";

	if ((fd = open_file_from_user (input_filename)) == (FILE*) 0)
		exit (1);

	U_stack_push_sinclude (incfile_malloc_c (
		  yy_create_buffer (fd, YY_BUF_SIZE)
		, input_filename
		, 0
		, fd
	));

	if (!flag_nointerfaces && !flag_cubix) {
		if ((fd = open_file (path_get_full (intrin))) == (FILE*) 0)
			exit (1);

		U_stack_push_sinclude (incfile_malloc_c (
			  yy_create_buffer (fd, YY_BUF_SIZE)
			, path_get_full (intrin)
			, 0
			, fd
		));
	}
	else
		++num_yywrap;

	include_yywrap ();
}

include_yywrap ()
{
	Incfile *incfile;

	switch (++num_yywrap) {
	case 1:
		flag_externals = 1;  /* to allow externals in lib/intrin.itf */
		flag_w = 1;
		break;

	case 2:
/* gets called after processing intrin.itf */
		if (!flag_nointerfaces) {
			intrin_set_intent ();
			table_user_to_intrin ();
		}

		zzlineno = 0;
		flag_externals = flag_externals_user;
		flag_w = flag_w_user;
		break;

	default:
		num_yywrap = 3;  /* to prevent overflow ... */
	}

	if (sinclude_top > 0) {
		if (current_fd != (FILE*) 0)
			fclose (current_fd);

		incfile = U_stack_pop_sinclude ();

		yy_switch_to_buffer (incfile->state);
		input_filename = incfile->name;
		current_fd = incfile->fd;
		zzlineno = incfile->lineno;
		form_old_line = incfile->old_line;
		form_old_len = incfile->old_len;

		return 0;
	}
	else
		return 1;
}

include_add (str)
char *str;
{
	char *file;
	FILE *fd;

	file = &str[1];
	file [strlen (file) - 1] = '\0';

	if ((fd = open_file_include (file)) == (FILE*) 0)
		return;

	U_stack_push_sinclude (
		incfile_malloc_c (YY_CURRENT_BUFFER, input_filename, zzlineno, current_fd));

	U_stack_push_sinclude ( incfile_malloc_c (
		  yy_create_buffer (fd, YY_BUF_SIZE)
		, file
		, 0
		, fd
	));

	current_fd = (FILE*) 0;
	include_yywrap ();
}

kind_check ()
{
	char *p;

	for (p = yytext; *p != '\0'; p++)
		if (*p == '_') {
			f90ni_kind ();
			return;
		}
}
